@@ -2,14 +2,17 @@ import json
 import logging
 import re
 import secrets
+from ipaddress import ip_address, ip_network
+from operator import attrgetter
 from time import time
 from typing import Any, AsyncGenerator, Awaitable, Callable, Dict, List, Optional
 
 from sentry_sdk import capture_event
 from sentry_sdk.utils import event_from_exception, exc_info_from_error
+from starlette.applications import Starlette
 from starlette.middleware.base import BaseHTTPMiddleware
 from starlette.requests import Request
-from starlette.responses import Response
+from starlette.responses import RedirectResponse, Response
 from starlette.routing import get_name as get_endpoint_name
 
 from . import glove
@@ -17,14 +20,21 @@ from .utils import get_ip
 
 logger = logging.getLogger('foxglove.middleware')
 request_logger = logging.getLogger('foxglove.bad_requests')
+CallNext = Callable[[Request], Awaitable[Response]]
 
-__all__ = 'ErrorMiddleware', 'CsrfMiddleware'
+__all__ = (
+    'ErrorMiddleware',
+    'CsrfMiddleware',
+    'HostRedirectMiddleware',
+    'CloudflareCheckMiddleware',
+    'request_log_extra',
+)
 
 
 class ErrorMiddleware(BaseHTTPMiddleware):
     def __init__(
         self,
-        app,
+        app: Starlette,
         should_warn: Callable[[Response], bool] = None,
         get_user: Callable[[Request], Awaitable[Dict[str, Any]]] = None,
     ):
@@ -36,7 +46,7 @@ class ErrorMiddleware(BaseHTTPMiddleware):
 
         self.glove = glove
 
-    async def dispatch(self, request: Request, call_next: Callable[[Request], Awaitable[Response]]) -> Response:
+    async def dispatch(self, request: Request, call_next: CallNext) -> Response:
         try:
             request.state.start_time = get_request_start(request)
 
@@ -57,36 +67,9 @@ class ErrorMiddleware(BaseHTTPMiddleware):
     async def log(
         self, request: Request, *, exc: Optional[Exception] = None, response: Optional[Response] = None
     ) -> None:
-        extra = dict(duration=f'{(time() - request.state.start_time) * 1000:0.2f}ms', query=dict(request.query_params))
-        if endpoint := request.scope.get('endpoint'):
-            extra.update(route_endpoint=get_endpoint_name(endpoint), path_params=dict(request.path_params))
-
-        if exc:
-            extra['exception_extra'] = exc_extra(exc)
-        else:
-            assert response is not None
-
-            extra.update(
-                response_status=response.status_code,
-                response_headers=dict(response.headers),
-                response_body=lenient_json(await self.response_body(response)),
-            )
-
-        view_ref = re.sub(r'\d{2,}', '{number}', str(request.url.path))
-        event_data = dict(
-            extra=extra,
-            user=await self.user_info(request),
-            transaction=view_ref,
-            request=dict(
-                url=str(request.url),
-                query_string=request.url.query,
-                cookies=dict(request.cookies),
-                headers=dict(request.headers),
-                method=request.method,
-                data=lenient_json(getattr(request, '_body', None)),
-                inferred_content_type=request.headers.get('Content-Type'),
-            ),
-        )
+        event_data = await request_log_extra(request, exc, response)
+        event_data['user'] = await self.user_info(request)
+        view_ref = event_data['transaction']
 
         if exc:
             level = 'error'
@@ -94,6 +77,7 @@ class ErrorMiddleware(BaseHTTPMiddleware):
             fingerprint = view_ref, request.method, repr(exc)
             request_logger.exception(message, extra=event_data)
         else:
+            assert response is not None
             level = 'warning'
             message = f'"{line_one(request)}", unexpected response: {response.status_code}'
             request_logger.warning(message, extra=event_data)
@@ -143,6 +127,56 @@ class ErrorMiddleware(BaseHTTPMiddleware):
             return b''.join(body_chunks)
 
 
+async def request_log_extra(
+    request: Request, exc: Optional[Exception] = None, response: Optional[Response] = None
+) -> Dict[str, Any]:
+    extra = dict(query=dict(request.query_params))
+
+    if start_time := getattr(request.state, 'start_time', None):
+        extra['duration'] = f'{(time() - start_time) * 1000:0.2f}ms'
+
+    if endpoint := request.scope.get('endpoint'):
+        extra.update(route_endpoint=get_endpoint_name(endpoint), path_params=dict(request.path_params))
+
+    if exc:
+        extra['exception_extra'] = exc_extra(exc)
+    elif response:
+        extra.update(
+            response_status=response.status_code,
+            response_headers=dict(response.headers),
+            response_body=lenient_json(await get_response_body(response)),
+        )
+
+    return dict(
+        extra=extra,
+        user=dict(ip_address=get_ip(request)),
+        transaction=re.sub(r'/\d{2,}/', '/{number}/', str(request.url.path)),
+        request=dict(
+            url=str(request.url),
+            query_string=request.url.query,
+            cookies=dict(request.cookies),
+            headers=dict(request.headers),
+            method=request.method,
+            data=lenient_json(request.scope.get('_body')),
+            inferred_content_type=request.headers.get('Content-Type'),
+        ),
+    )
+
+
+async def get_response_body(response: Response) -> bytes:
+    if hasattr(response, 'body'):
+        return response.body
+    else:
+        body_chunks = []
+        async for chunk in response.body_iterator:
+            if not isinstance(chunk, bytes):
+                chunk = chunk.encode(response.charset)
+            body_chunks.append(chunk)
+
+        response.body_iterator = async_gen_list(body_chunks)
+        return b''.join(body_chunks)
+
+
 async def async_gen_list(list_: List[bytes]) -> AsyncGenerator[bytes, None]:
     for c in list_:
         yield c
@@ -197,7 +231,7 @@ class CsrfMiddleware(BaseHTTPMiddleware):
     This prevents CSRF especially if the cookie has same_site strict
     """
 
-    async def dispatch(self, request: Request, call_next: Callable[[Request], Awaitable[Response]]) -> Response:
+    async def dispatch(self, request: Request, call_next: CallNext) -> Response:
         session_id = request.session.get(session_id_key)
         benign_request = request.method in {'HEAD', 'GET', 'OPTIONS'}
         if not benign_request and session_id is None:
@@ -209,3 +243,81 @@ class CsrfMiddleware(BaseHTTPMiddleware):
         if benign_request and response.status_code == 200 and session_id is None:
             request.session[session_id_key] = secrets.token_urlsafe()
         return response
+
+
+class HostRedirectMiddleware(BaseHTTPMiddleware):
+    def __init__(self, app: Starlette, host: str = None):
+        super().__init__(app)
+        self.host = host or glove.settings.domain
+        assert self.host, 'host must not be None in HostRedirectMiddleware'
+
+    async def dispatch(self, request: Request, call_next: CallNext) -> Response:
+        if request.url.hostname == self.host:
+            return await call_next(request)
+        else:
+            return RedirectResponse(request.url.replace(hostname=self.host), status_code=301)
+
+
+class IPRangeCounter:
+    __slots__ = 'range', 'counter'
+
+    def __init__(self, network_range: str):
+        self.range = ip_network(network_range)
+        self.counter = 0
+
+    def __repr__(self):
+        return f'IPRangeCounter({self.range}, {self.counter})'
+
+
+class CloudflareCheckMiddleware(BaseHTTPMiddleware):
+    default_response_body = b'Request incorrectly routed, this looks like a problem with your DNS or Proxy.'
+
+    def __init__(self, app: Starlette, response_text: str = None):
+        super().__init__(app)
+        self.response_body = response_text.encode() if response_text else self.default_response_body
+        # got from https://www.cloudflare.com/ips/
+        self.ip_ranges = [
+            IPRangeCounter('173.245.48.0/20'),
+            IPRangeCounter('103.21.244.0/22'),
+            IPRangeCounter('103.22.200.0/22'),
+            IPRangeCounter('103.31.4.0/22'),
+            IPRangeCounter('141.101.64.0/18'),
+            IPRangeCounter('108.162.192.0/18'),
+            IPRangeCounter('190.93.240.0/20'),
+            IPRangeCounter('188.114.96.0/20'),
+            IPRangeCounter('197.234.240.0/22'),
+            IPRangeCounter('198.41.128.0/17'),
+            IPRangeCounter('162.158.0.0/15'),
+            IPRangeCounter('104.16.0.0/12'),
+            IPRangeCounter('172.64.0.0/13'),
+            IPRangeCounter('131.0.72.0/22'),
+            IPRangeCounter('2400:cb00::/32'),
+            IPRangeCounter('2606:4700::/32'),
+            IPRangeCounter('2803:f800::/32'),
+            IPRangeCounter('2405:b500::/32'),
+            IPRangeCounter('2405:8100::/32'),
+            IPRangeCounter('2a06:98c0::/29'),
+            IPRangeCounter('2c0f:f248::/32'),
+        ]
+
+    async def dispatch(self, request: Request, call_next: CallNext) -> Response:
+        client = request.scope.get('client')
+        if client and self.find_network(client[0]):
+            return await call_next(request)
+
+        extra = {'extra': await request_log_extra(request)}
+        logger.warning('Request not routed through cloudflare client=%s url="%s"', client, request.url, extra=extra)
+        return Response(self.response_body, status_code=400)
+
+    def find_network(self, ip: str) -> bool:
+        try:
+            ip = ip_address(ip.strip())
+        except ValueError:
+            pass
+        else:
+            for r in self.ip_ranges:
+                if ip in r.range:
+                    r.counter += 1
+                    self.ip_ranges.sort(key=attrgetter('counter'), reverse=True)
+                    return True
+        return False
